# frozen_string_literal: true

require 'ougai'
require 'mono_logger'
require 'zuora_observability/logging/formatter'
require 'ougai/formatters/customizable'

module ZuoraObservability
  # A configurable logger that can be used for Rails and additional libraries
  class Logger < Ougai::Logger
    # TODO(hartley): enable silence method for Rails 5.2 asset requests
    # https://github.com/tilfin/ougai/wiki/Use-as-Rails-logger
    def initialize(logdev, **)
      super

      # NOTE(hartley): the purpose for the original split between Ougai and
      # MonoLogger was that MonoLogger enables logging in a trap context
      # https://github.com/tilfin/ougai/issues/74
      # By using our own Logger class, we can override the LogDevice created
      # by ruby with MonoLogger's, enabling logging in a trap context
      @logdev = MonoLogger::LocklessLogDevice.new(logdev)
    end

    def create_formatter
      return ZuoraObservability::Logging::Formatter.new if ZuoraObservability.configuration.json_logging

      formatter = Ougai::Formatters::Customizable.new(
        format_err: method(:custom_error_formatter),
        format_data: method(:custom_data_formatter),
        format_msg: method(:custom_message_formatter)
      )
      formatter.datetime_format = '%FT%T.%3NZ'
      formatter
    end

    def custom_message_formatter(severity, datetime, _progname, data)
      msg = data.delete(:msg)
      "#{severity.ljust(6)} #{datetime}: #{msg}"
    end

    def custom_data_formatter(data)
      data.delete_if do |k|
        %i[app_instance_id tenant_ids organization environment].include? k
      end

      return nil if data.blank?

      "#{'DATA'.ljust(6)} #{Time.current.strftime('%FT%T.%3NZ')}: #{data.to_json}"
    end

    def custom_error_formatter(data)
      return nil unless data.key?(:err)

      err = data.delete(:err)
      "  #{err[:name]} (#{err[:message]})\n #{err[:stack]}"
    end

    def self.custom_logger(name: '', level: Rails.logger.present? ? Rails.logger.level : MonoLogger::INFO, type: :ougai)
      if type == :ougai
        logger = new($stdout, level: level, progname: name)
      else
        ActiveSupport::Deprecation.warn(<<-MSG.squish)
          Creating a non-ougai custom_logger is deprecated and will be removed
          in a future version of zuora_observability.
        MSG

        require 'mono_logger'
        logger = MonoLogger.new(STDOUT)
        logger.level = level
        logger.formatter =  proc do |serverity, datetime, progname, msg|
          begin
            msg = JSON.parse(msg)
          rescue JSON::ParserError => ex
          end
          if ZuoraObservability.configuration.json_logging
            require 'json'
            store = {
              name: name,
              level: serverity,
              timestamp: datetime.strftime('%FT%T.%6NZ'),
              pid: Process.pid,
              message: name == "ActionMailer" ? msg.strip : msg
            }
            JSON.dump(store) + "\n"
          else
            format('%s %s: %s', serverity.ljust(6), datetime, msg) + "\n"
          end
        end
      end

      logger
    end
  end
end
