# frozen_string_literal: true

module ZuoraObservability
  module Logging
    # Custom Options for lograge to add additional fields to logged data
    module CustomOptions
      IGNORE_HEADERS = %w[
        RAW_POST_DATA
        REQUEST_METHOD
        REQUEST_URI
        REQUEST_PATH
        PATH_INFO
        CONTENT_TYPE
        ORIGINAL_FULLPATH
        QUERY_STRING
        HTTP_COOKIE
      ].freeze

      IGNORE_PARAMS = %w[controller action format].freeze

      FILTERED = '[FILTERED]'

      class << self
        def call(event)
          items = {
            msg: 'Rails Request',
            params: event.payload[:params].as_json(except: IGNORE_PARAMS).to_s,
            zuora_trace_id: event.payload[:headers]['HTTP_ZUORA_REQUEST_ID'],
            error: event.payload[:exception_object]
          }

          if event.payload[:headers].present?
            request_headers = filter_request_headers(event.payload[:headers])

            filtered_headers = filter_sensitive_headers(request_headers)

            items.merge!({ headers: filtered_headers })
          end

          items
        end

        private

        def filter_request_headers(action_dispatch_headers)
          action_dispatch_headers.env.select do |header|
            next false if IGNORE_HEADERS.include? header

            header =~ /^HTTP_/ || ActionDispatch::Http::Headers::CGI_VARIABLES.include?(header)
          end
        end

        def filter_sensitive_headers(headers)
          filtered = {}

          auth = headers['HTTP_AUTHORIZATION']
          filtered['HTTP_AUTHORIZATION'] = filter_auth_header(auth) if auth

          filtered['HTTP_API_TOKEN'] = FILTERED if headers['HTTP_API_TOKEN']

          headers.merge(filtered)
        end

        def filter_auth_header(authorization)
          return FILTERED unless authorization.include? 'Basic'

          encoded = authorization.split(' ', 2).second
          user, _password = Base64.decode64(encoded).split(':')

          "Basic #{user}:#{FILTERED}"
        end
      end
    end
  end
end
