# frozen_string_literal: true

module ZuoraObservability
  module Logging
    # Formats data into ZECS Logging Standard and dumps as JSON
    # https://tag.pages.gitlab.zeta.tools/standards/logging
    class Formatter < Ougai::Formatters::Bunyan
      ECS_VERSION = '1.5.0'

      ZECS_TOP_LEVEL_FIELDS = %i[
        error event http organization process url user
      ].freeze
      ECS_HTTP_FIELDS = %i[request response].freeze

      def _call(severity, time, progname, data)
        output = base_fields(severity, time, progname, data)

        other_fields = get_fields_for(ZECS_TOP_LEVEL_FIELDS, data)
        output.merge!(other_fields)

        zuora_fields = ZuoraFields.call(data)
        output[:zuora] = zuora_fields unless zuora_fields.empty?

        dump(output)
      end

      # Note(hartley): Ougai::Formatters::ForJson requires this be present
      def convert_time(data)
        data[:@timestamp] = data[:@timestamp].utc.iso8601(3)
      end

      private

      def get_fields_for(list, data)
        final_hash = {}

        list.each do |field|
          field_output = send("#{field}_fields", data)
          final_hash[field] = field_output unless field_output.empty?
        end

        final_hash
      end

      # Fields required by ECS
      def base_fields(severity, time, progname, data)
        {
          :@timestamp => time,
          message: data[:msg],
          ecs: { version: ECS_VERSION },
          log: { level: severity, logger: progname || @app_name },
          service: { name: Env.name, version: Env.version },
          trace: { id: data[:trace_id] },
          tags: ['zecs']
        }
      end

      # errors
      def error_fields(data)
        return {} unless data[:error] || data[:err]

        error = data[:err] || serialize_exc(data[:error])

        {
          message: error[:message],
          stack_trace: error[:stack],
          type: error[:name]
        }.compact
      end

      # http
      def http_fields(data)
        get_fields_for(ECS_HTTP_FIELDS, data)
      end

      # http.request
      def request_fields(data)
        request = data.slice(:method, :params).merge(data.fetch(:request, {}))

        {
          method: request[:method],
          body: ({ content: request[:params] } if request.key? :params)
        }.compact
      end

      # http.response
      def response_fields(data)
        response = data.slice(:status).merge(data.fetch(:response, {}))

        {
          status_code: response[:status],
          body: ({ content: response[:params] } if response.key? :params),
        }.compact
      end

      # organization
      def organization_fields(data)
        {
          name: data[:organization]
        }.compact
      end

      # process
      def process_fields(_data)
        { id: Process.pid }
      end

      # url
      def url_fields(data)
        {
          full: data.dig(:url, :full),
          path: data[:path]
        }.compact
      end

      # user
      def user_fields(data)
        {
          id: data[:user_id],
          email: data[:email]
        }.compact
      end

      # event
      def event_fields(data)
        full_action = "#{data[:controller]}##{data[:action]}"

        {
          action: (full_action unless full_action == '#'),
          duration: data[:duration]
        }.compact
      end
    end

    # The Zuora Extension to ECS defined by ZECS
    module ZuoraFields
      ZECS_VERSION = '1.1'

      class << self
        # zuora top level field
        def call(data)
          hash = base_fields(data)

          z_http = http_fields(data)
          hash[:http] = z_http unless z_http.empty?

          z_service = service_fields(data)
          hash[Env.name.to_sym] = z_service unless z_service.empty?

          return {} if hash.empty?

          hash.merge({ ecs: { version: ZECS_VERSION } })
        end

        def base_fields(data)
          {
            cp_id: data[:zuora_track_id],
            entity_ids: data[:entity_ids],
            environment: data[:environment],
            tenant_id: data[:tenant_ids],
            trace_id: data[:zuora_trace_id]
          }.compact
        end

        # zuora.http
        def http_fields(data)
          z_http_request = {
            headers: data[:headers] || data.dig(:request, :headers),
            headers_blob: data.dig(:request, :headers_blob)
          }.compact

          z_http_response = {
            headers: data.dig(:response, :headers),
            headers_blob: data.dig(:response, :headers_blob)
          }.compact

          {
            request: (z_http_request unless z_http_request.empty?),
            response: (z_http_response unless z_http_response.empty?)
          }.compact
        end

        # Service's Custom Fields
        def service_fields(data)
          {
            app_instance_id: data[:app_instance_id]
          }.merge(data.fetch(:zecs_service, {})).compact
        end
      end
    end
  end
end
